"""Interface to CurseForge.

This module contains definitons of classes wrapping project feeds, games
and any other resources available from the Curse network.
"""


import bz2
from contextlib import contextmanager
from datetime import datetime, timezone, timedelta
from functools import lru_cache
from io import BytesIO
from pathlib import Path
from typing import TextIO, Mapping, Sequence

import attr
import ijson
import requests
import sqlalchemy
from attr import validators as vld
from sqlalchemy import Column, Integer, String
from sqlalchemy import or_, bindparam
from sqlalchemy.orm.session import Session as SQLSession
from sqlalchemy.ext.baked import bakery
from sqlalchemy.ext.declarative import declarative_base

from .util import default_new_session, default_cache_dir

# Used exceptions -- make them available in this namespace
from requests.exceptions import HTTPError  # noqa: F401
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound  # noqa: F401


@attr.s(slots=True)
class Feed:
    """Interface to the Curse Project Feed for a particular game.

    The project feed is a set of bzip2-compressed json files that
    contains all of the add-ons for a game. The set consists of
    a complete feed (`complete.json.bz2`) and a hourly feed
    (`hourly.json.bz2`). Both feeds contain a timestamp of last change
    (in ms, for some reason), which is also accessible separately by
    appending a `.txt` suffix to the feed URL (i.e.
    `complete.json.bz2.txt`).
    """

    #: Base URL
    _BASEURL = 'http://clientupdate-v6.cursecdn.com/feed/addons/{id}/v10'
    #: Complete feed suffix
    _COMPLETE_URL = 'complete.json.bz2'

    #: Curse internal game identification
    game_id = attr.ib(validator=vld.instance_of(int))
    #: The :class:`requests.Session` to use for network requests
    session = attr.ib(
        validator=vld.optional(vld.instance_of(requests.Session)),
        default=None,
    )

    @property
    def complete_url(self) -> str:
        """Fully expanded URL of complete feed."""

        parts = (
            self._BASEURL.format(id=self.game_id),
            self._COMPLETE_URL,
        )

        return '/'.join(parts)

    @property
    def complete_timestamp_url(self) -> str:
        """Fully expanded URL of complete feed timestamp."""

        return self.complete_url + '.txt'

    @staticmethod
    @contextmanager
    def _decode_contents(feed: bytes) -> TextIO:
        """Decode the provided data from bz2 to text.

        The :arg:`feed` is assumed to be bz2-encoded text data in utf-8
        encoding.

        Keyword arguments:
            feed: The data to be decoded.

        Returns: Decoded text stream.
        """

        with BytesIO(feed) as compressed, \
                bz2.open(compressed, mode='rt', encoding='utf-8') as stream:
            yield stream

    @contextmanager
    def fetch_complete(self) -> TextIO:
        """Provide complete feed contents.

        Returns:
            Text stream of complete feed contents, that should be used
            in with-statement to be closed afterwards.

        Raises:
            requests.HTTPError: When an HTTP error occurs when fetching feed.
        """

        session = default_new_session(self.session)

        resp = session.get(self.complete_url)
        resp.raise_for_status()

        with self._decode_contents(resp.content) as text:
            yield text

    @staticmethod
    def _decode_timestamp(ms_timestamp: int) -> datetime:
        """Convert timestamp in ms into a :class:`datetime` object.

        Keyword arguments:
            ms_timestamp: The timestamp, in miliseconds, presumed to be in UTC.

        Returns:
            :class:`datetime` object pointing to the same point in time
            as the timestamp.
        """

        return datetime.fromtimestamp(ms_timestamp/1000, timezone.utc)

    def fetch_complete_timestamp(self) -> datetime:
        """Provide current complete feed time signature.

        .. note:: The timestamp is assumed to be in UTC (as it should be).

        Returns:
            Datetime object pointed to the same point in time as the timestamp.

        Raises:
            requests.HTTPError: When an HTTP error occurs while fetching.
        """

        session = default_new_session(self.session)

        resp = session.get(self.complete_timestamp_url)
        resp.raise_for_status()

        return self._decode_timestamp(int(resp.content))


@attr.s(slots=True)
class Database:
    """Interface to the local database of addons for a particular game.

    The addon database is stored in SQLite3 DB file. Its main purpose is to
    store feed's data locally and to prevent unnecessary re-downloading
    and re-parsing of them.

    The timestamp of the data can (and should) be stored directly
    in the DB file header (using `pragma user_version`). However, this only
    accepts 32 bits wide integer, so the timestamp should be stored
    fractionless (floored to whole second), and it is vulnerable
    to the "Year 2038" problem.
    """

    _SCHEME = 'sqlite://'  #: DB URI scheme.
    _BASENAME = '{game_name}-addons.sqlite'  #: DB URI basename format

    #: Name uniquely identifiyng the game.
    game_name = attr.ib(validator=vld.instance_of(str))
    #: Location of the database on the filesystem.
    root_dir = attr.ib(validator=vld.instance_of(Path))

    @property
    def uri(self) -> str:
        """Constructs full DB URI for this database."""

        return '/'.join((
            self._SCHEME,
            str(self.root_dir.resolve()),
            self._BASENAME.format(game_name=self.game_name),
        ))

    @property
    @lru_cache()
    def engine(self) -> sqlalchemy.engine.Engine:
        """Provide connection pool for the database."""

        return sqlalchemy.create_engine(self.uri)

    @property
    def version(self) -> datetime:
        """Read the version/timestamp of the data in the database."""

        timestamp, = self.engine.execute('PRAGMA user_version').first()
        return datetime.fromtimestamp(timestamp, timezone.utc)

    @version.setter
    def version(self, newver: datetime) -> None:
        """Set the version/timestamp of the data in the database."""

        # Cannot use SQL interpolation in PRAGMA statements :(
        # Force integral formating that the value is indeed an integer
        query = 'PRAGMA user_version = {:d}'.format(int(newver.timestamp()))
        with self.engine.begin() as conn:
            conn.execute(query)

    def session(self) -> SQLSession:
        """Create new session for batch database communication."""

        return SQLSession(bind=self.engine)


# Declarative base class for DB table definitions
AddonBase = declarative_base()
# Cache for pre-compiling SQL queries
SQLBakery = bakery()


class Mod(AddonBase):
    """Single game modification data adapter.

    This class serves as the adapter between JSON in the project feed
    and the local database.
    """

    __tablename__ = 'mods'

    #: Internal Curse mod identification
    id = Column(Integer, primary_key=True, autoincrement=False)
    #: Official mod name
    name = Column(String, index=True)
    #: Short mod description
    summary = Column(String, index=True)

    def __repr__(self) -> str:
        fmt = 'Mod(id={0.id!r}, name={0.name!r}, summary={0.summary!r})'
        return fmt.format(self)

    def __eq__(self, other: 'Mod') -> bool:
        if not isinstance(other, Mod):
            return NotImplemented

        partials = (
            self.id == other.id,
            self.name == other.name,
            self.summary == other.summary,
        )
        return all(partials)

    # Adapter methods

    @classmethod
    def from_json(cls, jobj: Mapping) -> 'Mod':
        """Construct new instance from JSON.

        Keyword arguments:
            jobj: The JSON data to use.

        Returns:
            New instance.
        """

        fields = 'id', 'name', 'summary'

        data = {k: jobj[k.capitalize()] for k in fields}
        return cls(**data)

    # Prepared queries

    @classmethod
    def search(cls, connection: SQLSession, term: str) -> Sequence['Mod']:
        """Search for Mods that contain TERM in name or summary.

        Keyword arguments:
            connection: Database connection to ask on.
            term: The term to search for.

        Returns:
            Sequence of matching mods (possibly empty).
        """

        query = SQLBakery(lambda conn: conn.query(cls))
        query += lambda q: q.filter(or_(
            cls.name.like(bindparam('term')),
            cls.summary.like(bindparam('term')),
        ))
        query += lambda q: q.order_by(cls.name)

        return query(connection).params(term='%{}%'.format(term)).all()

    @classmethod
    def find(cls, connection: SQLSession, name: str) -> 'Mod':
        """Find exactly one Mod named NAME.

        Keyword Arguments:
            connection: Database connection to ask on.
            name: The name of the mod to search for.

        Returns:
            The requested mod.

        Raises:
            NoResultsFound: The name does not match any known mod.
            MultipleResultsFound: The name is too ambiguous,
                multiple matching mods found.
        """

        query = SQLBakery(lambda conn: conn.query(cls))
        query += lambda q: q.filter(cls.name.like(bindparam('name')))

        return query(connection).params(name='%{}%'.format(name)).one()


@attr.s(init=False, slots=True)
class Game:
    """Interface to the projects related to one game in Curse network.

    The purpose of this class is to aggregate any objects providing related
    functionality and provide high-level procedural "glue" tying them
    together in one neat package.
    """

    # Primary attributes – must be supplied by user
    id = attr.ib(validator=vld.instance_of(int))  #: Curse internal game ID
    name = attr.ib(validator=vld.instance_of(str))  #: Human-readable name

    # Secondary/Derived attributes
    database = attr.ib(validator=vld.instance_of(Database))
    feed = attr.ib(validator=vld.instance_of(Feed))

    def __init__(
        self,
        id: int,
        name: str,
        *,
        session: requests.Session = None,
        cache_dir: Path = None
    ):
        """Initialize and create all the data for a game.

        Keyword arguments:
            id: Curse internal game identification.
            name: Human-readable name.
            session: :class:`requests.Session` to use for network calls.
            cache_dir: Path to the game's cache (which include mod database).
        """

        session = default_new_session(session)
        cache_dir = default_cache_dir(cache_dir)

        self.id = id
        self.name = name

        self.database = Database(game_name=name.lower(), root_dir=cache_dir)
        self.feed = Feed(game_id=id, session=session)

        # Fill the database, if it does not exists
        epoch = datetime.fromtimestamp(0, tz=timezone.utc)
        if self.database.version == epoch:
            AddonBase.metadata.create_all(self.database.engine)

    def refresh_data(self):
        """Download, store and index fresh version of the game add-ons."""

        sess = self.database.session()

        # Destroy indexes and truncate old data
        sess.query(Mod).delete()

        # Parse the feed's data
        # TODO: Extract feed's timestamp from the JSON
        with self.feed.fetch_complete() as feed:
            addons = ijson.items(feed, 'data.item')
            mods = filter(
                lambda a: a['CategorySection']['Path'] == 'mods',
                addons,
            )

            sess.add_all(Mod.from_json(m) for m in mods)

        sess.commit()

        # Write the timestamp
        self.database.version = self.feed.fetch_complete_timestamp()

    def have_fresh_data(
        self,
        valid_period: timedelta = timedelta(hours=24),
        *,
        now: datetime = datetime.now(tz=timezone.utc)
    ) -> bool:
        """Check if the data in the database are still fresh (enough).

        By default, the data are considered to be actual for 24 hours
        after the publication of the project feed.

        Keyword arguments:
            valid_period: How much time has to pass since the data import
                for the data to be considered stale.
            now: Specify the point in time to be considered 'now'.

        Returns:
            True if the data are still considered fresh, False otherwise.
        """

        time_passed = now - self.database.version
        return time_passed < valid_period



"""Various utilities and language enhancements."""


from datetime import datetime
from functools import partial
from pathlib import Path

import requests
import yaml
import xdg.BaseDirectory
from iso8601 import parse_date

try:
    from yaml import CLoader as YAMLLoader, CDumper as YAMLDumper
except ImportError:
    from yaml import Loader as YAMLLoader, Dumper as YAMLDumper


# Consistent names definitions
RESOURCE_NAME = __package__


# Filesystem standard directories
def default_cache_dir(directory: Path = None) -> Path:
    """Get default cache dir if directory is None,
    else return directory as it is.
    """

    if directory is None:
        return Path(xdg.BaseDirectory.save_cache_path(RESOURCE_NAME))
    else:
        return directory


def default_data_dir(directory: Path = None) -> Path:
    """Get default data dir if directory is None,
    else return directory as it is.
    """

    if directory is None:
        return Path(xdg.BaseDirectory.save_data_path(RESOURCE_NAME))
    else:
        return directory


def default_new_session(session: requests.Session = None) -> requests.Session:
    """Create new Requests' Session, if none is provided.
    Otherwise, return session as it is.
    """

    if session is None:
        return requests.Session()
    else:
        return session


# YAML tweaks

# Force YAML to use ISO-8601 standard for timestamps
def yaml_timestamp_representer(dumper, data):
    """Custom representer for datetime objects in YAML."""
    return dumper.represent_scalar('!!timestamp', data.isoformat())
YAMLDumper.add_representer(datetime, yaml_timestamp_representer)


def yaml_timestamp_constructor(loader, node):
    """Custom constructor for datetime objects from YAML."""
    value = loader.construct_scalar(node)
    return parse_date(value)
YAMLLoader.add_constructor('!!timestamp', yaml_timestamp_constructor)


yamlload = partial(yaml.load, Loader=YAMLLoader)
yamlload.__doc__ = """Load and parse an YAML document."""


yamldump = partial(yaml.dump, Dumper=YAMLDumper, default_flow_style=False)
yamldump.__doc__ = """Dump and store an YAML document."""



"""Global test configuration"""


import os

import betamax


# Ensure cassete dir
CASSETE_DIR = 'tests/cassetes/'
if not os.path.exists(CASSETE_DIR):
    os.makedirs(CASSETE_DIR)

record_mode = 'none' if os.environ.get('TRAVIS_BUILD') else 'once'

with betamax.Betamax.configure() as config:
    config.cassette_library_dir = CASSETE_DIR
    config.default_cassette_options.update({
        'record_mode': record_mode,
        'preserve_exact_body_bytes': True,
    })



"""Interactive text user interface parts."""

import curses
from typing import Callable, Sequence, Iterable, Optional

import urwid

from .curse import Mod


#: Signature of valid menu callbacks
ModItemCallback = Callable[[Mod, urwid.Button], None]


class ModMenu(urwid.ListBox):
    """Menu presenting a choice from a list of :class:`Mod`s.

    The menu remembers which mod was chosen, in its :attr:`chosen`
    attribute (which defaults to None).
    """

    __slots__ = 'chosen',

    #: Color palette of the mod menu
    palette = [
        ('title',
            'dark green,bold', 'black', 'bold', '#0a0,bold', '#000'),
        ('title_focus',
            'white', 'dark green', 'underline', '#ff8', '#0a0'),
        ('description',
            'light gray', 'black'),
    ]

    class Item(urwid.Pile):
        """Display widget for a single :class:`Mod`."""

        def __init__(self, mod: Mod, *callbacks: Iterable[ModItemCallback]):
            """Wrap mod in the set of display widgets.

            Keyword arguments:
                mod: The :class:`Mod` to be wrapped.
                callbacks: The functions to be called when this object
                    is selected.
            """

            btn_prefix = '  ◈ '

            # Construct button (the selectable part)
            btn = urwid.Button('')
            btn._w = urwid.AttrMap(
                urwid.SelectableIcon([btn_prefix, mod.name], 2),
                'title', 'title_focus',
            )
            for callback in callbacks:
                urwid.connect_signal(btn, 'click', callback, user_args=[mod])

            # Construct the mod summary
            text = urwid.Padding(
                urwid.AttrMap(urwid.Text(mod.summary), 'description'),
                left=len(btn_prefix)*2,
            )

            pile = btn, text
            super().__init__(pile)

    def __init__(self, choices: Sequence[Mod]):
        """Create menu for choices.

        Keyword arguments:
            choices: The :class:`Mod`s to choose from.
        """

        items = [self.Item(m, self.choose, self.end_loop) for m in choices]
        super().__init__(urwid.SimpleFocusListWalker(items))

        self.chosen = None

    def choose(self, mod: Mod, btn: urwid.Button) -> None:
        """Record the choice.

        Keyword arguments:
            mod: The mod to record as the last choice.
            btn: Unused, exists for signature compatibility.
        """

        self.chosen = mod

    def end_loop(self, *unused) -> None:
        """End main event loop.

        Raises:
            urwid.ExitMainLoop: Every time it is called.
        """

        raise urwid.ExitMainLoop()


def exit_loop_on_q_esc(key: str):
    """End urwid.MainLoop on keypress of 'q' or 'esc'."""

    if key in {'q', 'Q', 'esc'}:
        raise urwid.ExitMainLoop()


def select_mod(
    choices: Sequence[Mod],
    header: Optional[str] = None,
    footer: Optional[str] = None,
) -> Optional[Mod]:
    """Present user with a TUI menu and return his choice.

    Keyword arguments:
        choices: The :class:`Mod`s to choose from.
        header: Optional menu heading.
        footer: Optional menu footing.

    Returns:
        The selected mod.
    """

    menu = ModMenu(choices)

    head = urwid.Text(('title', header), align='center') if header else None
    foot = urwid.Text(('description', footer), align='center') if footer else None  # noqa: E501

    if head or foot:
        top = urwid.Frame(menu, head, foot)
    else:
        top = menu

    try:
        colors = curses.tigetnum('colors')
    except curses.error:  # Uninitialized terminal
        colors = 16

    event_loop = urwid.MainLoop(
        top,
        palette=ModMenu.palette,
        unhandled_input=exit_loop_on_q_esc,
    )
    event_loop.screen.set_terminal_properties(colors=colors)
    event_loop.run()

    return menu.chosen



"""Mod-pack file format interface."""

import re
from copy import deepcopy
from datetime import datetime
from enum import Enum, unique
from functools import total_ordering
from typing import Any, Callable, TextIO, Union

import cerberus
from iso8601 import parse_date

from .util import yamlload, yamldump, YAMLLoader, YAMLDumper


@unique
@total_ordering
class Release(Enum):
    """Enumeration of the possible release types of a mod file."""

    Alpha = 1
    Beta = 2
    Release = 4

    # Make the releases comparable
    def __is_same_enum(self: 'Release', other: Any) -> bool:
        """Detect if the compared value is of the same class."""
        return other.__class__ is self.__class__

    def __eq__(self: 'Release', other: 'Release') -> bool:
        if self.__is_same_enum(other):
            return self.value == other.value
        else:
            return NotImplemented

    def __ne__(self: 'Release', other: 'Release') -> bool:
        if self.__is_same_enum(other):
            return self.value != other.value
        else:
            return NotImplemented

    def __lt__(self: 'Release', other: 'Release') -> bool:
        if self.__is_same_enum(other):
            return self.value < other.value
        else:
            return NotImplemented

    # Nicer serialization to YAML
    @classmethod  # In order to not mix with enumeration members
    def yaml_tag(cls) -> str:
        return '!release'

    @classmethod
    def from_yaml(cls, loader, node) -> 'Release':
        """Constructs release from an YAML node."""

        name = loader.construct_scalar(node)
        return cls[name]

    @classmethod
    def to_yaml(cls, dumper, release):
        """Serialize release to an YAML node."""

        return dumper.represent_scalar(cls.yaml_tag(), release.name)

YAMLDumper.add_representer(Release, Release.to_yaml)
YAMLLoader.add_constructor(Release.yaml_tag(), Release.from_yaml)
YAMLLoader.add_implicit_resolver(
    Release.yaml_tag(),
    re.compile('^{}$'.format('|'.join(name for name in Release.__members__))),
    None,
)


# Custom cerberus validators and coercers
def valid_release(field: str, value: Any, error: Callable) -> bool:
    """Validate Release type."""
    if isinstance(value, Release):
        return True
    else:
        error(field, "Not a valid release: '{!s}'".format(value))


def isodate(date: Union[str, datetime]) -> datetime:
    """Convert ISO-8601 date to datetime, if needed."""

    if isinstance(date, datetime):
        return date
    else:
        return parse_date(date)


def to_release(release: Union[str, Release]) -> Release:
    """Convert Release name to Release, if necessary."""

    if isinstance(release, Release):
        return release
    else:
        return Release[release]


# Mod file schema
cerberus.schema_registry.add('mod-file', {
    'id': {'type': 'integer', 'required': True, 'coerce': int},
    'name': {'type': 'string', 'required': True},
    'date': {'type': 'datetime', 'required': True, 'coerce': isodate},
    'release': {'validator': valid_release, 'required': True, 'coerce': to_release},  # noqa: E501
    'dependencies': {'type': 'list', 'schema': {'type': 'integer'}},
})

# Mod schema
cerberus.schema_registry.add('mod', {
    'id': {'type': 'integer', 'required': True, 'coerce': int},
    'name': {'type': 'string'},
    'file': {'type': 'dict', 'schema': 'mod-file', 'required': True},
})

# Game schema
cerberus.schema_registry.add('game', {
    'name': {'type': 'string'},
    'version': {'type': 'string'},
})

# Pack file schema
cerberus.schema_registry.add('pack', {
    'game': {'type': 'dict', 'schema': 'game', 'required': True},
    'mods': {'type': 'list', 'schema': {'type': 'dict', 'schema': 'mod'}},
    'dependencies': {'type': 'list', 'schema': {'type': 'dict', 'schema': 'mod'}},  # noqa: E501
})


class ValidationError(ValueError):
    """Exception for reporting invalid pack data."""

    __slots__ = 'errors',

    def __init__(self, msg: str, errors: dict):
        super().__init__(msg)
        self.errors = errors


class ModPack:
    """Interface to single mod-pack data."""

    __slots__ = 'data',

    def __init__(self, data: dict):
        """Validate data for the mod-pack.

        Raises:
            ValidationError: If the data are not valid mod-pack data.
        """

        vld = cerberus.Validator(cerberus.schema_registry.get('pack'))
        # Do NOT change input data during validation!
        if not vld.validate(deepcopy(data)):
            raise ValidationError('Invalid pack data', vld.errors)

        self.data = vld.document

    @classmethod
    def create(cls, name: str, version: str) -> 'ModPack':
        """Create new, empty mod-pack.

        Keyword arguments:
            name: Name of the game to make pack for.
            version: Version of the game to make pack for.

        Returns:
            Newly created mod-pack.
        """

        return cls({
            'game': {
                'name': name,
                'version': version,
            },
        })

    @classmethod
    def from_yaml(cls, stream: TextIO) -> 'ModPack':
        """Load mod-pack data from a file stream.

        Keyword arguments:
            stream: Input YAML stream.

        Returns:
            Mod-pack loaded form the stream.

        Raises:
            ValidationError: If the stream does not contain valid pack data.
        """

        return cls(yamlload(stream))

    def to_yaml(self, stream: TextIO) -> None:
        """Serialize and save the mod-pack data to YAML stream.

        Keyword arguments:
            stream: The YAML stream to write to.
        """

        yamldump(self.data, stream)



"""Tests for the pack submodule"""

from copy import deepcopy
from datetime import datetime
from io import StringIO

import cerberus
import pytest

from mccurse import pack
from mccurse.util import yamldump, yamlload


@pytest.fixture
def pack_validator() -> cerberus.Validator:
    return cerberus.Validator(cerberus.schema_registry.get('pack'))


@pytest.fixture
def valid_mod_file() -> dict:
    return {
        'id': '42',
        'name': 'test-mod-file.jar',
        'date': '2017-01-24T18:01+01:00',
        'release': 'Beta',
    }


@pytest.fixture
def invalid_mod_file() -> dict:
    return {
        'name': 'testfile.zip',
        'date': 'yesterday',
        'release': 'Released',
        'deps': [123, 456],
    }


@pytest.fixture
def empty_mod() -> dict:
    """Empty (without file) mod entry"""
    return {
        'id': 74072,
        'name': 'Tinkers Contruct',
    }


@pytest.fixture
def valid_mod(empty_mod, valid_mod_file) -> dict:
    return dict(**empty_mod, file=deepcopy(valid_mod_file))


@pytest.fixture
def invalid_mod(empty_mod, invalid_mod_file) -> dict:
    return dict(**empty_mod, file=invalid_mod_file)


@pytest.fixture
def minimal_pack() -> dict:
    return {
        'game': {'version': '1.10.2'},
    }


@pytest.fixture
def valid_pack(valid_mod) -> dict:
    return {
        'game': {
            'name': 'Minecraft',
            'version': '1.10.2',
        },
        'mods': [deepcopy(valid_mod)],
        'dependencies': [],
    }


@pytest.fixture
def invalid_pack(invalid_mod) -> dict:
    return {
        'game': {
            'name': 'Minecraft',
            'version': '1.10.2',
        },
        'mods': [deepcopy(invalid_mod)],
        'dependencies': [],
    }


@pytest.fixture
def valid_yaml(valid_pack) -> StringIO:
    stream = StringIO()
    yamldump(valid_pack, stream)
    stream.seek(0)

    return stream


@pytest.fixture
def invalid_yaml(invalid_pack) -> StringIO:
    stream = StringIO()
    yamldump(invalid_pack, stream)
    stream.seek(0)

    return stream


def test_release():
    """Test release creation and ordering"""

    A = pack.Release['Alpha']
    B = pack.Release['Beta']
    R = pack.Release['Release']

    assert A == pack.Release['Alpha']
    assert A < B < R
    assert R > B > A
    assert A != B


def test_release_and_yaml():
    """Serialization of Release to YAML works as intended?"""

    data = [pack.Release['Alpha']]
    text = '- !release Alpha\n'

    assert yamldump(data) == text
    assert yamlload(text) == data


def test_mod_file_schema(valid_mod_file, invalid_mod_file):
    """Mod file schema bahving as expected?"""

    schema = cerberus.schema_registry.get('mod-file')
    vld = cerberus.Validator(schema)
    inv = cerberus.Validator(schema)

    assert vld.validate(valid_mod_file)
    assert isinstance(vld.document['id'], int)
    date = vld.document['date']
    assert isinstance(date, datetime) and date.tzinfo is not None
    assert isinstance(vld.document['release'], pack.Release)

    assert not inv.validate(invalid_mod_file)
    assert inv.errors


def test_mod_schema(valid_mod, invalid_mod):
    """Mod schema behaving as expected?"""

    schema = cerberus.schema_registry.get('mod')
    vld = cerberus.Validator(schema)
    inv = cerberus.Validator(schema)

    assert vld.validate(valid_mod)
    assert isinstance(vld.document['id'], int)
    assert isinstance(vld.document['file']['id'], int)

    assert not inv.validate(invalid_mod)


def test_pack_schema(minimal_pack, valid_pack, invalid_pack):
    """Pack schema behaving as expected?"""

    schema = cerberus.schema_registry.get('pack')
    minimal, valid, invalid = map(cerberus.Validator, [schema]*3)

    assert minimal.validate(minimal_pack)
    assert valid.validate(valid_pack)
    assert not invalid.validate(invalid_pack)


def test_modpack_init(valid_pack, invalid_pack):
    """ModPack.__init__ behaves as described?"""

    mp = pack.ModPack(valid_pack)
    assert mp.data

    with pytest.raises(pack.ValidationError):
        mp = pack.ModPack(invalid_pack)


def test_modpack_create():
    """Does ModPack.create work as expected?"""

    name, version = 'Minecraft', '1.10.2'
    mp = pack.ModPack.create(name, version)

    assert mp.data['game']['name'] == name
    assert mp.data['game']['version'] == version


def test_modpack_load(pack_validator, valid_yaml, valid_pack, invalid_yaml):
    """Loading from stream works as advertised?"""

    mp = pack.ModPack.from_yaml(valid_yaml)
    assert mp.data == pack_validator.normalized(valid_pack)

    with pytest.raises(pack.ValidationError):
        mp = pack.ModPack.from_yaml(invalid_yaml)


def test_modpack_dump(pack_validator, valid_pack):
    """Dumping to stream works as advertised?"""

    stream = StringIO()
    mp = pack.ModPack(valid_pack)
    expect = pack_validator.normalized(deepcopy(valid_pack))

    mp.to_yaml(stream)
    print(stream.getvalue())
    data = yamlload(stream.getvalue())

    assert data == expect



"""Tests for util submodule."""


from pathlib import Path

import requests
import xdg

from mccurse import util


def test_cache_use_default():
    """Use cache dir when no dir specified?"""

    INPUT = None
    EXPECT = Path(xdg.BaseDirectory.save_cache_path(util.RESOURCE_NAME))

    assert util.default_cache_dir(INPUT) == EXPECT


def test_cache_use_input():
    """Use existing dir when specified?"""

    INPUT = Path.home()
    EXPECT = INPUT

    assert util.default_cache_dir(INPUT) == EXPECT


def test_data_use_default():
    """Use data dir when no dir specified?"""

    INPUT = None
    EXPECT = Path(xdg.BaseDirectory.save_data_path(util.RESOURCE_NAME))

    assert util.default_data_dir(INPUT) == EXPECT


def test_data_use_input():
    """Use existing dir when specified?"""

    INPUT = Path.home()
    EXPECT = INPUT

    assert util.default_data_dir(INPUT) == EXPECT


def test_use_existing_session():
    """Use existing session when specified?"""

    INPUT = requests.Session()
    EXPECT = INPUT

    assert util.default_new_session(INPUT) is EXPECT


def test_make_new_session():
    """Make new session when none provided?"""

    INPUT = None

    assert isinstance(util.default_new_session(INPUT), requests.Session)



"""Interface to the Curse.RestProxy service."""

from typing import TextIO

import attr
import requests
from attr import validators as vld
from requests.auth import AuthBase

from .util import default_new_session, yamlload, yamldump


HOME_URL = 'https://curse-rest-proxy.azurewebsites.net/api'


@attr.s(slots=True)
class Authorization(AuthBase):
    """Authorization mechanism for the proxy service."""

    #: User identification number
    user_id = attr.ib(validator=vld.instance_of(int))
    #: Session token
    token = attr.ib(validator=vld.instance_of(str))

    def __call__(self, req: requests.Request) -> requests.Request:
        """Make the request authenticated."""

        header_fmt = 'Token {user_id}:{token}'
        req.headers['Authorization'] = header_fmt.format_map(attr.asdict(self))

        return req

    @classmethod
    def login(
        cls,
        username: str,
        password: str,
        *,
        session: requests.Session = None
    ) -> 'Authorization':
        """Login into the RestProxy service.

        Keyword arguments:
            username: Name of the user to log in.
            password: Password of the user to log in.
            session: :class:`requests.Session` to use for logging in.

        Returns:
            Authorization for the specified user.

        Raises:
            requests.HTTPError: On invalid credentials.
        """

        session = default_new_session(session)

        url = '/'.join((HOME_URL, 'authenticate'))
        resp = session.post(url, json={
            'username': username,
            'password': password,
        })

        resp.raise_for_status()

        data = resp.json()
        return cls(
            user_id=data['session']['user_id'],
            token=data['session']['token'],
        )

    @classmethod
    def load(cls, file: TextIO) -> 'Authorization':
        """Load stored credentials from file.

        Keyword arguments:
            file: Open YAML text stream to read from.

        Returns:
            Authorization previously saved to the file.

        Raises:
            ValueError: When the stream does not contain expected data.
        """

        data = yamlload(file)

        if not data or 'user_id' not in data or 'token' not in data:
            msg = 'Invalid authorization data: {!r}'.format(data)
            raise ValueError(msg)

        return cls(**data)

    def dump(self, file: TextIO) -> None:
        """Store credentials for future use.

        Keyword arguments:
            file: Open YAML text stream to write to.
        """

        yamldump(attr.asdict(self), file)



"""A setup module for the mccurse package."""

from distutils.command.build import build as orig_build
from pathlib import Path
from setuptools import setup, find_packages
from setuptools.command.develop import develop as orig_develop
from typing import Generator, TextIO


# Custom commands

class build(orig_build):
    sub_commands = [
        # Commands before original build
        ('compile_catalog', None),
    ] + orig_build.sub_commands + [
        # Commands after original build
    ]


class develop(orig_develop):
    def run(self):
        self.run_command('compile_catalog')
        super().run()


# Helper functions

def extract_deps(fd: TextIO) -> Generator[str, None, None]:
    """Extract dependencies from file."""
    yield from (
        line for line in fd
        if line and not line.startswith(('#', 'git+'))
    )


# Setup values preparation

here = Path(__file__).resolve().parent

# Long description
readme = here / 'README.rst'
with readme.open(encoding='utf-8') as rdm:
    long_description = rdm.read()


deps = here / 'dependencies.txt'
test_deps = here / 'test-dependencies.txt'

with deps.open(encoding='utf-8') as d:
    install_requires = list(extract_deps(d))
with test_deps.open(encoding='utf-8') as td:
    test_requires = list(extract_deps(td))
    # XXX add missing dependency specified with git URL
    test_requires += ['pyfakefs>2.9']

setup_requires = [
    'setuptools_scm',
    'babel',
]


# Setup call

setup(
    name='mccurse',

    # Take version automatically from the git tag
    use_scm_version=True,

    description='Minecraft Curse CLI Client',
    long_description=long_description,

    url='https://github.com/khardix/mccurse',

    author='Jan "Khardix" Staněk',
    author_email='khardix@gmail.com',

    license='AGPLv3+',
    classifiers=[
        # Development status
        'Development Status :: 3 - Alpha',

        # Target audience
        'Intended Audience :: End Users/Desktop',
        'Topic :: Games/Entertainment',
        'Topic :: Utilities',

        # License
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',  # noqa: E501

        # Python versions
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='minecraft modding automation',

    cmdclass={
        'build': build,
        'develop': develop,
    },

    packages=find_packages(exclude=('tests', 'docs')),
    package_data={
        'mccurse': ['locales/*/LC_MESSAGES/*.mo'],
    },

    # Requirements
    install_requires=install_requires,
    extras_require={
        'test': test_requires,
    },
    setup_requires=setup_requires,

    # Scripts entry points
    entry_points={
        'console_scripts': [
            'mccurse=mccurse.cli:cli',
        ],
    },
)



"""Package command line interface."""

import curses
from pathlib import Path

import click

from . import _
from .curse import Game, Mod
from .pack import ModPack
from .proxy import Authorization
from .tui import select_mod
from .util import default_data_dir


# Static data
MINECRAFT = {'id': 432, 'name': 'Minecraft'}


def check_minecraft_dir(root: Path) -> None:
    """Checks if the directory is a suitable space for minecraft mods.

    Keyword Arguments:
        root: The checked dir – should be a root dir of minecraft profile.

    Raises:
        FileNotFoundError: When some expected file or directory is not found.
    """

    mods_dir = root / 'mods'

    if not mods_dir.is_dir():
        raise FileNotFoundError(str(mods_dir))


@click.group()
@click.pass_context
def cli(ctx):
    """Minecraft Curse CLI client."""

    # Initialize terminal for querying
    curses.setupterm()

    # Add contextual values
    ctx.obj = {
        # Data directory
        'datadir': default_data_dir(),
        # Authentication file
        'authfile': default_data_dir() / 'token.yaml',
    }


@cli.command()
@click.option(
    '--refresh', is_flag=True, default=False,
    # NOTE: Help for refresh flag
    help=_('Force refreshing of search data.')
)
@click.argument('text', nargs=-1, type=str)
def search(refresh, text):
    """Search for TEXT in mods on CurseForge."""

    if not text:
        raise SystemExit(_('No text to search for!'))

    mc = Game(**MINECRAFT)

    text = ' '.join(text)
    refresh = refresh or not mc.have_fresh_data()

    if refresh:
        click.echo(_('Refreshing search data, please wait…'), err=True)
        mc.refresh_data()

    found = Mod.search(mc.database.session(), text)

    title = _('Search results for "{text}"').format(text=text)
    instructions = _(
        'Choose mod to open its project page, or press [q] to quit.'
    )

    chosen = select_mod(found, title, instructions)
    if chosen is not None:
        project_url_fmt = 'https://www.curseforge.com/projects/{mod.id}/'
        click.launch(project_url_fmt.format(mod=chosen))


@cli.command()
@click.option(
    '--user', '-u', prompt=_('Curse user name or email'),
    help=_('Curse user name or email')+'.',
)
@click.password_option(
    help=_('Curse password')+'.',
)
@click.pass_obj
def auth(ctx, user, password):
    """Authenticate user in Curse network."""

    token = Authorization.login(user, password)

    with ctx['authfile'].open(mode='w', encoding='utf-8') as file:
        token.dump(file)


@cli.command()
@click.option(
    '--profile', '-p', type=click.Path(exists=True, file_okay=False),
    default='.',
    help=_('Root profile directory')+'.',
)
@click.argument('version')
def new(profile, version):
    """Create a new modpack for Minecraft VERSION."""

    mc = Game(**MINECRAFT)
    profile = Path(str(profile))

    try:
        check_minecraft_dir(profile)
    except FileNotFoundError as err:
        msg = _('Profile directory is not valid: Missing {!s}').format(err)
        raise SystemExit(msg) from None

    modpack_file = profile / 'modpack.yaml'
    with modpack_file.open(mode='w', encoding='utf-8') as stream:
        ModPack.create(name=mc.name, version=version).to_yaml(stream)
